{% if is_eol %}
{# Creates a banner at the top of the page for EOL versions. #}
<div id='banner' class='Admonition caution'>
  <p>You are reading an unmaintained version of the Ansible documentation. Unmaintained Ansible versions can contain unfixed security vulnerabilities (CVE). Please upgrade to a maintained version. See <a href="/ansible/latest/">the latest Ansible documentation</a>.</p>
</div>
{% else %}
  <script>
    function startsWith(str, needle) {
      return str.slice(0, needle.length) == needle
    }
    function startsWithOneOf(str, needles) {
      return needles.some(function (needle) {
        return startsWith(str, needle);
      });
    }
    var banner = '';
    var survey_banner =
      '<div id="latest_survey_banner_id" class="admonition important">' +
      '<br>' +
      '<p>' +
        'Please take our ' +
        '<a href="https://www.surveymonkey.co.uk/r/2VWVJY3">survey</a> ' +
        'to help us improve support for collections and roles in GalaxyNG.' +
      '</p>' +
      '<br>' +
      '</div>';
    // Create a banner if we're not on the official docs site
    if (location.host == "docs.testing.ansible.com") {
      document.write('<div id="testing_banner_id" class="admonition important">' +
                     '<p>This is the testing site for Ansible Documentation. Unless you are reviewing pre-production changes, please visit the <a href="https://docs.ansible.com/ansible/latest/">official documentation website</a>.</p> <p></p>' +
                     '</div>');
    }
    {% if available_versions is defined %}
      // Create a banner
      current_url_path = window.location.pathname;

      var important = false;
      var msg = '<p>';
      if (startsWith(current_url_path, "/ansible-core/")) {
        msg += 'You are reading documentation for Ansible Core, which contains no plugins except for those in ansible.builtin. For documentation of the Ansible package, go to <a href="/ansible/latest">the latest documentation</a>.';
      } else if (startsWithOneOf(current_url_path, ["/ansible/latest/", "/ansible/{{ latest_version }}/"])) {
        /* temp banner to advertise GalaxyNG survey */
      //  banner += survey_banner;

        msg += 'You are reading the latest community version of the Ansible documentation. Red Hat subscribers, select <b>2.9</b> in the version selection to the left for the most recent Red Hat release.';
      } else if (startsWith(current_url_path, "/ansible/2.9/")) {
        msg += 'You are reading the latest Red Hat released version of the Ansible documentation. Community users can use this version, or select <b>latest</b> from the version selector to the left for the most recent community version.';
      } else if (startsWith(current_url_path, "/ansible/devel/")) {
        /* temp banner to advertise GalaxyNG survey */
      //  banner += survey_banner;

        /* temp banner to advertise survey
        important = true;
        msg += 'Please take our <a href="https://www.surveymonkey.co.uk/r/B9V3CDY">Docs survey</a> before December 31 to help us improve Ansible documentation.';
        */

        msg += 'You are reading the <b>devel</b> version of the Ansible documentation - this version is not guaranteed stable. Use the version selection to the left if you want the latest stable released version.';
      } else {
        msg += 'You are reading an older version of the Ansible documentation. Use the version selection to the left if you want the latest stable released version.';
      }
      msg += '</p>';

      banner += '<div id="banner_id" class="admonition ';
      banner += important ? 'important' : 'caution';
      banner += '">';
      banner += important ? '<br>' : '';
      banner += msg;
      banner += important ? '<br>' : '';
      banner += '</div>';
      document.write(banner);
    {% endif %}
  </script>
{% endif %}
